import os
from pathlib import Path
from typing import Iterator, Literal, TypeVar
import json
from logging import getLogger

from .task_queue import TaskQueue

logger = getLogger(__name__)
Bot = TypeVar("Bot")


class Config:
    """A class for handling configuration settings for the aiomql package.

    Attributes:
        record_trades (bool): Whether to keep record of trades or not.
        trade_record_mode: How to save trade, json or csv. Defaults to json
        filename (str): Name of the config file
        records_dir (str): Path to the directory where trade records are saved
        login (int): Trading account number
        password (str): Trading account password
        server (str): Broker server
        path (str): Path to terminal file
        timeout (int): Timeout for terminal connection
        _initialize (bool): First time initialization flag
        state (dict): A global state dictionary for storing data across the framework
        root_dir (str): The root directory of the project
    Notes:
        By default, the config class looks for a file named aiomql.json.
        You can change this by passing the filename and/or the config_dir keyword argument(s) to the constructor
        or the load_config method.
        By passing reload=True to the load_config method, you can reload and search again for the config file.
    """
    login: int = 0
    trade_record_mode: Literal['csv', 'json'] = 'csv'
    password: str = ""
    server: str = ""
    path: str | Path = ""
    timeout: int = 60000
    record_trades: bool = True
    filename: str = "aiomql.json"
    _initialize = True
    state: dict = {}
    root: Path
    root_dir: Path
    records_dir: Path
    config_dir: str = ''
    task_queue: TaskQueue = TaskQueue()
    bot: Bot = None
    _instance: 'Config'

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, **kwargs):
        reload = kwargs.pop('reload', False)
        self.load_config(reload=reload, **kwargs)

    def set_root(self, *, root: str | Path):
        root = Path(root) if str else root
        self.root = root.absolute().resolve()
        self.root_dir = self.root

    def __setattr__(self, key, value):
        if key == 'path':
            value = str(self.root_dir / Path(value).absolute().resolve())
        super().__setattr__(key, value)

    def set_attributes(self, **kwargs):
        """Set keyword arguments as object attributes

        Keyword Args:
            **kwargs: Object attributes and values as keyword arguments
        """
        [setattr(self, key, value) for key, value in kwargs.items()]

    @staticmethod
    def walk_to_root(path: str | Path) -> Iterator[str]:
        if not os.path.exists(path):
            raise IOError("Starting path not found")

        if os.path.isfile(path):
            path = os.path.dirname(path)

        last_dir = None
        current_dir = os.path.abspath(path)
        while last_dir != current_dir:
            yield current_dir
            parent_dir = os.path.abspath(os.path.join(current_dir, os.path.pardir))
            last_dir, current_dir = current_dir, parent_dir

    def find_config(self):
        try:
            path = self.root_dir / self.config_dir
            for dirname in self.walk_to_root(path):
                check_path = os.path.join(dirname, self.filename)
                if os.path.isfile(check_path):
                    return check_path
            return None
        except Exception as _:
            return

    def create_records_dir(self, *, records_dir: str | Path = 'records'):
        """Create records directory if it does not exist. By default, it is relative to the root directory of the
         project unless an absolute path is provided.

        Keyword Args:
            records_dir (str|Path): The directory to save trade records. Default is 'records'
        """
        try:
            if isinstance(records_dir, str):
                records_dir = self.root_dir / records_dir
            elif isinstance(records_dir, Path):
                records_dir = records_dir.absolute().resolve()
            records_dir.mkdir(parents=True, exist_ok=True)
            self.records_dir = records_dir
        except Exception as err:
            logger.warning(f"{err}: Unable to create records directory")

    def load_config(self, *, file: str = None, reload: bool = True, filename: str = None,
                    config_dir: str = '', **kwargs):
        """Load configuration settings from a file.
        Keyword Args:
            file (str): The path to the file to load. If not provided, the file is searched for
            reload (bool): Whether to reload the config object. Default is True
            filename (str): The name of the file to load. If not provided, the default filename is used
            config_dir (str): The name of the directory to search for the file. Default is the root directory
            root_dir (str): The root directory of the project
            kwargs: Additional keyword arguments
        """
        if not (self._initialize or reload):
            return
        data = {}
        self.filename = filename or self.filename
        self.config_dir = config_dir or self.config_dir
        root_dir = kwargs.pop('root_dir', None)
        records_dir = kwargs.pop('records_dir', 'records')
        if self._initialize or (root_dir is not None):
            self.set_root(root=(root_dir or '.'))
            self.create_records_dir(records_dir=records_dir)

        if (file := (file or self.find_config())) is None:
            logger.warning("No Config File Found")
        else:
            fh = open(file, mode="r")
            data = json.load(fh)
            fh.close()
        data |= kwargs
        self.set_attributes(**data)
        self._initialize = False

    def account_info(self) -> dict[str, int | str]:
        """Returns Account login details as found in the config object if available

        Returns:
            dict: A dictionary of login details
        """
        return {"login": self.login, "password": self.password, "server": self.server}



from datetime import datetime
import asyncio
from logging import getLogger
from typing import Callable

import MetaTrader5
from MetaTrader5 import BookInfo, SymbolInfo, AccountInfo, Tick, TerminalInfo, TradeOrder, TradeDeal, \
    TradePosition, OrderSendResult, OrderCheckResult

from .constants import TimeFrame, CopyTicks, OrderType
from .errors import Error
from .config import Config

logger = getLogger()


class BaseMeta(type):
    def __new__(mcs, cls_name, bases, cls_dict):
        defaults = MetaTrader5.__dict__
        defaults = {f'_{key}': value for key, value in defaults.items() if not key.startswith('_')}
        cls_dict |= defaults
        return super().__new__(mcs, cls_name, bases, cls_dict)


class MetaTrader(metaclass=BaseMeta):
    _account_info: Callable
    _copy_rates_from: Callable
    _copy_rates_from_pos: Callable
    _copy_rates_range: Callable
    _copy_ticks_from: Callable
    _copy_ticks_range: Callable
    _history_deals_get: Callable
    _history_deals_total: Callable
    _history_orders_get: Callable
    _history_orders_total: Callable
    _initialize: Callable
    _last_error: Callable
    _login: Callable
    _market_book_add: Callable
    _market_book_get: Callable
    _market_book_release: Callable
    _order_calc_margin: Callable
    _order_calc_profit: Callable
    _order_check: Callable
    _order_send: Callable
    _orders_get: Callable
    _orders_total: Callable
    _positions_get: Callable
    _positions_total: Callable
    _shutdown: Callable
    _symbol_info: Callable
    _symbol_info_tick: Callable
    _symbol_select: Callable
    _symbols_get: Callable
    _symbols_total: Callable
    _terminal_info: Callable
    _version: Callable
    error: Error
    config: Config

    def __init__(self):
        self.config = Config()

    async def __aenter__(self) -> 'MetaTrader':
        """
        Async context manager entry point.
        Initializes the connection to the MetaTrader terminal.

        Returns:
            MetaTrader: An instance of the MetaTrader class.
        """
        await self.initialize(**Config().account_info())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Async context manager exit point. Closes the connection to the MetaTrader terminal.
        """
        await self.shutdown()

    async def login(self, login: int, password: str, server: str, timeout: int = 60000) -> bool:
        """
        Connects to the MetaTrader terminal using the specified login, password and server.

        Args:
            login (int): The trading account number.
            password (str): The trading account password.
            server (str): The trading server name.
            timeout (int): The timeout for the connection in seconds.

        Returns:
            bool: True if successful, False otherwise.
        """
        return await asyncio.to_thread(self._login, login, password=password, server=server, timeout=timeout)

    async def initialize(self, path: str = "", login: int = 0, password: str = "", server: str = "",
                         timeout: int | None = None, portable=False) -> bool:
        """
        Initializes the connection to the MetaTrader terminal. All parameters are optional.

        Keyword Args:
            path (str): The path to the MetaTrader terminal executable.
            login (int): The trading account number.
            password (str): The trading account password.
            server (str): The trading server name.
            timeout (int): The timeout for the connection in seconds.
            portable (bool): If True, the terminal will be launched in portable mode.

        Returns:
            bool: True if successful, False otherwise.
        """
        args = (str(path),) if path else ()
        kwargs = {key: value for key, value in (('login', login), ('password', password), ('server', server),
                                                ('timeout', timeout), ('portable', portable)) if value}
        return await asyncio.to_thread(self._initialize, *args, **kwargs)

    async def shutdown(self) -> None:
        """
        Closes the connection to the MetaTrader terminal.

        Returns:
            None: None
        """
        return await asyncio.to_thread(self._shutdown)

    async def last_error(self) -> tuple[int, str]:
        try:
            return await asyncio.to_thread(self._last_error)
        except Exception as err:
            logger.warning(f'Error in obtaining last error.')
            return -1, str(err)

    async def version(self) -> tuple[int, int, str] | None:
        """"""
        res = await asyncio.to_thread(self._version)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining version information.{self.error.description}')
        return res

    async def account_info(self) -> AccountInfo | None:
        """"""
        res = await asyncio.to_thread(self._account_info)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining account information.{self.error.description}')
        return res

    async def terminal_info(self) -> TerminalInfo | None:
        res = await asyncio.to_thread(self._terminal_info)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining terminal information.{self.error.description}')
            return res
        return res

    async def symbols_total(self) -> int:
        return await asyncio.to_thread(self._symbols_total)

    async def symbols_get(self, group: str = "") -> tuple[SymbolInfo] | None:
        kwargs = {'group': group} if group else {}
        res = await asyncio.to_thread(self._symbols_get, **kwargs)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining symbols.{self.error.description}')
            return res
        return res

    async def symbol_info(self, symbol: str) -> SymbolInfo | None:
        res = await asyncio.to_thread(self._symbol_info, symbol)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining information for {symbol}.{self.error.description}')
            return res
        return res

    async def symbol_info_tick(self, symbol: str) -> Tick | None:
        res = await asyncio.to_thread(self._symbol_info_tick, symbol)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining tick for {symbol}.{self.error.description}')
            return res
        return res

    async def symbol_select(self, symbol: str, enable: bool) -> bool:
        return await asyncio.to_thread(self._symbol_select, symbol, enable)

    async def market_book_add(self, symbol: str) -> bool:
        return await asyncio.to_thread(self._market_book_add, symbol)

    async def market_book_get(self, symbol: str) -> tuple[BookInfo] | None:
        res = await asyncio.to_thread(self._market_book_get, symbol)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining market depth content for {symbol}.{self.error.description}')
            return res
        return res

    async def market_book_release(self, symbol: str) -> bool:
        return await asyncio.to_thread(self._market_book_release, symbol)

    async def copy_rates_from(self, symbol: str, timeframe: TimeFrame, date_from: datetime | float, count: int):
        res = await asyncio.to_thread(self._copy_rates_from, symbol, timeframe, date_from, count)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining rates for {symbol}.{self.error.description}')
            return res
        return res

    async def copy_rates_from_pos(self, symbol: str, timeframe: TimeFrame, start_pos: int, count: int):
        res = await asyncio.to_thread(self._copy_rates_from_pos, symbol, timeframe, start_pos, count)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining rates for {symbol}.{self.error.description}')
            return res
        return res

    async def copy_rates_range(self, symbol: str, timeframe: TimeFrame, date_from: datetime | float,
                               date_to: datetime | float):
        res = await asyncio.to_thread(self._copy_rates_range, symbol, timeframe, date_from, date_to)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining rates for {symbol}.{self.error.description}')
            return res
        return res

    async def copy_ticks_from(self, symbol: str, date_from: datetime | float, count: int, flags: CopyTicks):
        res = await asyncio.to_thread(self._copy_ticks_from, symbol, date_from, count, flags)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining ticks for {symbol}.{self.error.description}')
            return res
        return res

    async def copy_ticks_range(self, symbol: str, date_from: datetime | float, date_to: datetime | float,
                               flags: CopyTicks):
        res = await asyncio.to_thread(self._copy_ticks_range, symbol, date_from, date_to, flags)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining ticks for {symbol}.{self.error.description}')
            return res
        return res

    async def orders_total(self) -> int:
        return await asyncio.to_thread(self._orders_total)

    async def orders_get(self, group: str = "", ticket: int = 0, symbol: str = "") -> tuple[TradeOrder] | None:
        """Get active orders with the ability to filter by symbol or ticket. There are three call options.
           Call without parameters. Return active orders on all symbols

        Keyword Args:
            symbol (str): Symbol name. Optional named parameter. If a symbol is specified, the ticket parameter is ignored.

            group (str): The filter for arranging a group of necessary symbols. Optional named parameter. If the group is specified, the function
                returns only active orders meeting a specified criteria for a symbol name.

            ticket (int): Order ticket (ORDER_TICKET). Optional named parameter.

        Returns:
            tuple[TradeOrder]: A list of active trade orders as TradeOrder objects
        """
        kwargs = {key: value for key, value in (('group', group), ('ticket', ticket), ('symbol', symbol)) if value}
        res = await asyncio.to_thread(self._orders_get, **kwargs)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining orders.{self.error.description}')
            return res
        return res

    async def order_calc_margin(self, action: OrderType, symbol: str, volume: float, price: float) -> float | None:
        res = await asyncio.to_thread(self._order_calc_margin, action, symbol, volume, price)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in calculating margin.{self.error.description}')
            return res
        return res

    async def order_calc_profit(self, action: OrderType, symbol: str, volume: float, price_open: float,
                                price_close: float) -> float | None:
        res = await asyncio.to_thread(self._order_calc_profit, action, symbol, volume, price_open, price_close)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in calculating profit.{self.error.description}')
            return res
        return res

    async def order_check(self, request: dict) -> OrderCheckResult:
        return await asyncio.to_thread(self._order_check, request)

    async def order_send(self, request: dict) -> OrderSendResult:
        return await asyncio.to_thread(self._order_send, request)

    async def positions_total(self) -> int:
        return await asyncio.to_thread(self._positions_total)

    async def positions_get(self, group: str = "", ticket: int = None, symbol: str = "") -> tuple[TradePosition] | None:
        kwargs = {key: value for key, value in (('group', group), ('ticket', ticket), ('symbol', symbol)) if value}
        res = await asyncio.to_thread(self._positions_get, **kwargs)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in obtaining open positions.{self.error.description}')
            return res
        return res

    async def history_orders_total(self, date_from: datetime | float, date_to: datetime | float) -> int:
        return await asyncio.to_thread(self._history_orders_total, date_from, date_to)

    async def history_orders_get(self, date_from: datetime | float = None, date_to: datetime | float = None,
                                 group: str = '', ticket: int = None, position: int = None) -> tuple[TradeOrder] | None:
        kwargs = {key: value for key, value in (('group', group), ('ticket', ticket), ('position', position)) if value}
        args = tuple(arg for arg in (date_from, date_to) if arg)
        res = await asyncio.to_thread(self._history_orders_get, *args, **kwargs)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in getting orders.{self.error.description}')
            return res
        return res

    async def history_deals_total(self, date_from: datetime | float, date_to: datetime | float) -> int:
        return await asyncio.to_thread(self._history_deals_total, date_from, date_to)

    async def history_deals_get(self, date_from: datetime | float = None, date_to: datetime | float = None,
                                group: str = '', ticket: int = None, position: int = None) -> tuple[TradeDeal] | None:
        kwargs = {key: value for key, value in (('group', group), ('ticket', ticket), ('position', position)) if value}
        args = tuple(arg for arg in (date_from, date_to) if arg)
        res = await asyncio.to_thread(self._history_deals_get, *args, **kwargs)
        if res is None:
            err = await self.last_error()
            self.error = Error(*err)
            logger.warning(f'Error in getting deals.{self.error}')
            return res
        return res



import asyncio
from datetime import time, timedelta, datetime
from asyncio import sleep, iscoroutinefunction
from typing import Literal, Callable
from logging import getLogger

from .positions import Positions

logger = getLogger(__name__)


def delta(obj: time) -> timedelta:
    """Get the timedelta of a datetime.time object.

    Args:
        obj (datetime.time): A datetime.time object.
    """
    return timedelta(hours=obj.hour, minutes=obj.minute, seconds=obj.second, microseconds=obj.microsecond)


class Session:
    """A session is a time period between two datetime.time objects specified in utc.

    Attributes:
        start (datetime.time): The start time of the session.
        end (datetime.time): The end time of the session.
        on_start (str): The action to take when the session starts. Default is None.
        on_end (str): The action to take when the session ends. Default is None.
        custom_start (Callable): A custom function to call when the session starts. Default is None.
        custom_end (Callable): A custom function to call when the session ends. Default is None.
        name (str): A name for the session. Default is a combination of start and en
    """
    def __init__(self, *, start: int | time, end: int | time,
                 on_start: Literal['close_all', 'close_win', 'close_loss', 'custom_start'] = None,
                 on_end: Literal['close_all', 'close_win', 'close_loss', 'custom_end'] = None,
                 custom_start: Callable = None, custom_end: Callable = None, name: str = ''):
        """Create a session.

        Keyword Args:
            start (int | datetime.time): The start time of the session in UTC.
            end (int | datetime.time): The end time of the session in UTC.
            on_start (Literal['close_all', 'close_win', 'close_loss', 'custom_start']): The action to take when the
                session starts. Default is None.
            on_end (Literal['close_all', 'close_win', 'close_loss', 'custom_end']): The action to take when the session
                ends. Default is None.
            custom_start (Callable): A custom function to call when the session starts. Default is None.
            custom_end (Callable): A custom function to call when the session ends. Default is None.
            name (str): A name for the session. Default is a combination of start and end.
        """
        self.start = start if isinstance(start, time) else time(hour=start)
        self.end = end if isinstance(end, time) else time(hour=end)
        self.on_start = on_start
        self.on_end = on_end
        self.custom_start = custom_start
        self.custom_end = custom_end
        self.name = name or f'{self.start} - {self.end}'

    def __contains__(self, item: time):
        if self.start > self.end:
            m1 = time(hour=23, minute=59, second=59, microsecond=9999)
            m2 = time(hour=0)
            return self.start <= item <= m1 or m2 <= item < self.end
        return self.start <= item < self.end

    def __str__(self):
        return f'{self.start}-->{self.name}-->{self.end}' if self.name else f'{self.start}-->{self.end}'

    def __repr__(self):
        return f'{self.start}-->{self.end}'

    def __len__(self):
        return (delta(self.start) - delta(self.end)).seconds

    async def begin(self):
        """Call the action specified in on_start or custom_start."""
        await self.action(self.on_start)

    async def close(self):
        """Call the action specified in on_end or custom_end."""
        await self.action(self.on_end)

    async def action(self, action):
        """Used by begin and close to call the action specified.

        Args:
            action (Literal['close_all', 'close_win', 'close_loss', 'custom_start', 'custom_end']): The action to take.
        """
        try:
            position = Positions()
            positions = await position.positions_get()

            match action:
                case 'close_all':
                    await asyncio.gather(*(position.close(price=pos.price_current, ticket=pos.ticket,
                                                          order_type=pos.type, volume=pos.volume,
                                                          symbol=pos.symbol) for pos in positions),
                                         return_exceptions=True)

                case 'close_win':
                    await asyncio.gather(
                        *(position.close(price=pos.price_current, ticket=pos.ticket, order_type=pos.type,
                                         volume=pos.volume, symbol=pos.symbol) for pos in positions if pos.profit > 0),
                        return_exceptions=True)

                case 'close_loss':
                    await asyncio.gather(
                        *(position.close(price=pos.price_current, ticket=pos.ticket, order_type=pos.type,
                                         volume=pos.volume, symbol=pos.symbol) for pos in positions if
                          pos.profit < 0), return_exceptions=True)

                case 'custom_end':
                    if iscoroutinefunction(self.custom_end):
                        await self.custom_end()
                    self.custom_end()

                case 'custom_start':
                    if iscoroutinefunction(self.custom_start):
                        await self.custom_start()
                    self.custom_start()

                case _:
                    pass
        except Exception as exe:
            logger.warning(f'Failed to call action {action} due to {exe}')

    def until(self):
        """Get the seconds until the session starts from the current time in seconds."""
        return (delta(self.start) - delta(datetime.utcnow().time())).seconds


class Sessions:
    """Sessions allow you to run code at specific times of the day. It is a collection of Session objects.
    Sessions are sorted by start time. The sessions object is an asynchronous context manager.

    Attributes:
        sessions (list[Session]): A list of Session objects.
        current_session (Session): The current session.

    Methods:
        find: Find a session that contains a datetime.time object.
        find_next: Find the next session that contains a datetime.time object.
        check: Check if the current session has started and if not, wait until it starts.
    """
    def __init__(self, *sessions: Session):
        self.sessions = list(sessions)
        self.sessions.sort(key=lambda x: (x.start, x.end))
        self.current_session = None

    def find(self, obj: time) -> Session | None:
        """Find a session that contains a datetime.time object.

        Args:
            obj (datetime.time): A datetime.time object.

        Returns:
            Session | None: A Session object or None if not found.
        """
        for session in self.sessions:
            if obj in session:
                return session
        return None

    def find_next(self, obj: time) -> Session:
        """Find the next session that contains a datetime.time object.

        Args:
            obj (datetime.time): A datetime.time object.

        Returns:
            Session: A Session object.
        """
        for session in self.sessions:
            if obj < session.start:
                return session
        return self.sessions[0]

    def __contains__(self, item: time):
        return True if self.find(item) is not None else False

    async def __aenter__(self):
        await self.check()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.current_session.close()

    async def check(self):
        """Check if the current session has started and if not, wait until it starts."""
        now = datetime.utcnow().time()
        current_session = self.find(now)
        if current_session:
            if self.current_session:
                if self.current_session == current_session:
                    return
                await self.current_session.close()

            self.current_session = current_session
            await self.current_session.begin()
            return

        await self.current_session.close() if self.current_session else ...
        current_session = self.find_next(now)
        secs = current_session.until() + 10
        logger.info(f'sleeping for {secs} seconds until next {current_session} session')
        await sleep(secs)
        self.current_session = current_session
        await self.current_session.begin()



"""Utility functions for aiomql."""

import decimal
from .candle import Candles, Candle


def dict_to_string(data: dict, multi=False) -> str:
    """Convert a dict to a string. Useful for logging.

    Args:
        data (dict): The dict to convert.
        multi (bool, optional): If True, each key-value pair will be on a new line. Defaults to False.

    Returns:
        str: The string representation of the dict.
    """
    sep = '\n' if multi else ', '
    return f"{sep}".join(f"{key}: {value}" for key, value in data.items())


def round_off(value: float, step: float, round_down: bool = False) -> float:
    """Round off a number to the nearest step."""
    with decimal.localcontext() as ctx:
        ctx.rounding = decimal.ROUND_DOWN if round_down else decimal.ROUND_UP
        return float(decimal.Decimal(str(value)).quantize(decimal.Decimal(str(step))))


def find_bearish_fractal(candles: Candles) -> Candle | None:
    for i in range(len(candles) - 3, 1, -1):
        if candles[i].high > max(candles[i - 1].high, candles[i + 1].high, candles[i - 2].high, candles[i + 2].high):
            return candles[i]


def find_bullish_fractal(candles: Candles) -> Candle | None:
    for i in range(len(candles) - 3, 1, -1):
        if candles[i].low < min(candles[i - 1].low, candles[i + 1].low, candles[i - 2].low, candles[i + 2].low):
            return candles[i]



import asyncio
from datetime import datetime
from logging import getLogger

from pandas import DataFrame
import pandas as pd

from .core.config import Config
from .core.meta_trader import MetaTrader, CopyTicks, OrderType
from .core.models import TradeDeal, TradeOrder

logger = getLogger(__name__)


class History:
    """The history class handles completed trade deals and trade orders in the trading history of an account.

    Attributes:
        deals (list[TradeDeal]): Iterable of trade deals
        orders (list[TradeOrder]): Iterable of trade orders
        total_deals: Total number of deals
        total_orders (int): Total number orders
        group (str): Filter for selecting history by symbols.
        ticket (int): Filter for selecting history by ticket number
        position (int): Filter for selecting history deals by position
        mt5 (MetaTrader): MetaTrader instance
        config (Config): Config instance
    """
    mt5: MetaTrader
    config: Config

    def __init__(self, *, date_from: datetime | int = None, date_to: datetime | int = None,
                 group: str = "", ticket: int = None, position: int = None):
        """
        Args:
            date_from (datetime, float): Date the orders are requested from. Set by the 'datetime' object or as a
                number of seconds elapsed since 1970.01.01. Defaults to twenty-four hours from the current time in 'utc'

            date_to (datetime, float): Date up to which the orders are requested. Set by the 'datetime' object or as a
                number of seconds elapsed since 1970.01.01. Defaults to the current time in "utc"

            group (str): Filter for selecting history by symbols.
            ticket (int): Filter for selecting history by ticket number
            position (int): Filter for selecting history deals by position
        """
        self.config = Config()
        self.mt5 = MetaTrader()
        self.date_from = date_from
        self.date_to = date_to
        self.group = group
        self.ticket = ticket
        self.position = position
        self.deals: list[TradeDeal] = []
        self.orders: list[TradeOrder] = []
        self.total_deals: int = 0
        self.total_orders: int = 0

    async def init(self, deals=True, orders=True):
        """Get history deals and orders

        Keyword Args:
            deals (bool): If true get history deals during initial request to terminal
            orders (bool): If true get history orders during initial request to terminal
        """
        self.deals = await self.get_deals() if deals else tuple()
        self.orders = await self.get_orders() if orders else tuple()
        self.total_deals = len(self.deals)
        self.total_orders = len(self.orders)

    async def get_deals(self, *, date_from: datetime | int = None, date_to: datetime | int = None, group: str = '',
                        retries: int = 3) -> tuple[TradeDeal, ...]:
        """Get deals from trading history using the parameters set in the constructor.

        Returns:
            tuple[TradeDeal]: A list of trade deals
        """
        if retries < 1:
            logger.warning(f'Failed to get deals: {self.mt5.error}')
            return tuple()

        date_from, date_to, group = date_from or self.date_from, date_to or self.date_to, group or self.group
        deals = await self.mt5.history_deals_get(date_from=date_from, date_to=date_to, group=group)

        if deals is not None:
            return tuple(TradeDeal(**deal._asdict()) for deal in deals)

        if self.mt5.error.is_connection_error():
            await asyncio.sleep(retries)
            return await self.get_deals(date_from=date_from, date_to=date_to, group=group, retries=retries-1)

        logger.warning(f'Failed to get deals: {self.mt5.error}')
        return tuple()

    async def get_deals_ticket(self, *, ticket: int = None) -> tuple[TradeDeal, ...]:
        """Call specifying the order ticket. Return all deals having the specified order ticket in the DEAL_ORDER
        property.

        Args:
            ticket (int): The order ticket

        Returns:
            tuple[TradeDeal]: A tuple of all deals with the order ticket
        """
        ticket = ticket or self.ticket
        assert ticket is not None, 'ticket not provided'
        deals = await self.mt5.history_deals_get(ticket=ticket)
        return tuple(sorted([TradeDeal(**deal._asdict()) for deal in deals or []], key=lambda x: x.time_msc))

    async def get_deals_position(self, *, position: int = None) -> tuple[TradeDeal, ...]:
        """
        Get all deals with the specified position ticket in the DEAL_POSITION_ID property
        Args:
            position (int): The position ticket

        Returns:
            tuple[TradeDeal]: A tuple of all deals with the position ticket
        """
        position = position or self.position
        assert position is not None, 'position not provided'
        deals = await self.mt5.history_deals_get(position=position)
        return tuple(sorted([TradeDeal(**deal._asdict()) for deal in deals or []], key=lambda x: x.time_msc))

    async def deals_total(self, *, date_from: int | datetime = None, date_to: int | datetime = None) -> int:
        """Get total number of deals within the specified period in the constructor.
        Args:
            date_from (int|datetime): Date the orders are requested from. Set by the 'datetime' object or as a number of
                seconds elapsed since 1970.01.01.
            date_to (int|datetime): Date up to which the orders are requested. Set by the 'datetime' object or as a
                number of seconds elapsed since 1970.01.01.
        Returns:
            int: Total number of Deals
        """
        date_from, date_to = date_from or self.date_from, date_to or self.date_to
        assert date_from is not None and date_to is not None, 'date_from and/or date_to not provided'
        total_deals = await self.mt5.history_deals_total(date_from, date_to)
        return total_deals

    async def get_orders(self, *, date_from: datetime | int = None, date_to: datetime | int = None, group: str = '',
                         retries: int = 3) -> tuple[TradeOrder, ...]:
        """Get orders from trading history using the parameters set in the constructor or the method arguments.

        Returns:
            list[TradeOrder]: A list of trade orders
        """
        if retries < 1:
            logger.warning(f'Failed to get orders: {self.mt5.error}')
            return tuple()

        date_from, date_to, group = date_from or self.date_from, date_to or self.date_to, group or self.group
        orders = await self.mt5.history_orders_get(date_from=date_from, date_to=date_to, group=group)
        if orders is not None:
            return tuple(TradeOrder(**order._asdict()) for order in orders)

        if self.mt5.error.is_connection_error():
            await asyncio.sleep(retries)
            return await self.get_orders(date_from=date_from, date_to=date_to, group=group, retries=retries - 1)

        logger.warning(f'Failed to get orders: {self.mt5.error}')
        return tuple()

    async def get_order_ticket(self, ticket: int | None = None) -> TradeOrder | None:
        ticket = ticket or self.ticket
        assert isinstance(ticket, int), 'ticket not provided'
        orders = await self.mt5.history_orders_get(ticket=ticket)
        if orders and (order := orders[0]).ticket == ticket:
            return TradeOrder(**order._asdict())
        return None

    async def get_orders_position(self, position: int = None) -> tuple[TradeOrder, ...]:
        """
        Call specifying the position ticket. Return all orders with a position ticket specified in the
        ORDER_POSITION_ID property

        Args:
            position: The position ticket

        Returns:
            tuple[TradeOrder]: A tuple of all orders with the position ticket
        """
        position = position or self.position
        assert isinstance(position, int), 'position not provided'
        orders = await self.mt5.history_orders_get(position=position)
        return tuple(sorted([TradeOrder(**order._asdict()) for order in orders or []], key=lambda x: x.time_done_msc))

    async def orders_total(self, date_from: int | datetime = None, date_to: int | datetime = None) -> int:
        """Get total number of orders within the specified period in the constructor.

        Returns:
            int: Total number of orders
        """
        date_from, date_to = date_from or self.date_from, date_to or self.date_to
        assert date_from is not None and date_to is not None, 'date_from and/or date_to not provided'
        total_orders = await self.mt5.history_orders_total(date_from, date_to)
        return total_orders

    async def track_order(self, *, position: int = None, end_time: datetime = None) -> DataFrame:
        """
        Track an order from the time it was opened to the time it was closed or any given time.
         The tracking is done by getting the ticks
        for the order symbol from the time the order was opened to the time it was closed. The profit for each tick is
        calculated using the order type, symbol, initial volume, open price and the bid or ask price of the tick
        depending on the order type.
        Args:
            end_time (datetime): The time to stop tracking the order. If not provided, the tracking will continue until
                the order is closed.
            position (int): The position ticket
            end_time (int): The time to stop tracking the order in seconds. If not provided, the tracking will continue
             until the order is closed.
        Returns:
            DataFrame: A pandas DataFrame of the ticks and profit for the order.
        """
        orders = await self.get_orders_position(position=position)
        deals = await self.get_deals_position(position=position)
        open_order = orders[0]
        open_deal = deals[0]
        close_deal = deals[-1]
        time_done = datetime.timestamp(end_time) if end_time is not None else close_deal.time
        time_done_msc = int(time_done * 1000)
        open_order.set_attributes(time_done_msc=time_done_msc, time_done=time_done, price_open=open_deal.price)
        ticks = await self.mt5.copy_ticks_range(open_order.symbol, open_order.time_setup, open_order.time_done,
                                                CopyTicks.ALL)
        data = pd.DataFrame(ticks)
        profit = lambda x: self.mt5._order_calc_profit(open_order.type, open_order.symbol, open_order.volume_initial,
                                                       open_order.price_open,
                                                       x.ask if open_order.type == OrderType.BUY else x.bid)
        data['profits'] = data.apply(profit, axis=1)
        data['time'] = pd.to_datetime(data['time'], unit='s')
        data.set_index('time', inplace=True)
        return data



"""Risk Assessment and Management"""
from .account import Account
from .positions import Positions


class RAM:
    account: Account
    risk_to_reward: float
    risk: float
    points: float
    pips: float
    min_amount: float = 0
    max_amount: float = 0
    risk_level: float = 50
    loss_limit: int = 3
    open_limit: int = 6

    def __init__(self, *, risk_to_reward: float = 1, risk: float = 0.01, **kwargs):
        """Initialize Risk Assessment and Management with the provided keyword arguments.

        Keyword Args:
            risk_to_reward (float): Risk to reward ratio. Defaults to 1
            risk (float): Percentage of account balance to risk per trade 0.01 # 1%
            kwargs: extra keyword arguments are set as object attributes
        """
        self.risk_to_reward = risk_to_reward
        self.risk = risk
        self.account = Account()
        [setattr(self, key, value) for key, value in kwargs.items()]

    async def get_amount(self) -> float:
        """Calculate the amount to risk per trade as a percentage of equity.

        Returns:
            float: Amount to risk per trade
        """
        await self.account.refresh()
        amount = self.account.margin_free * self.risk
        if self.min_amount and self.max_amount:
            return max(self.min_amount, min(self.max_amount, amount))
        return amount

    async def check_losing_positions(self, *, symbol: str = '') -> bool:
        """Check if the number of losing positions is greater than or equal the loss limit.

        Args:
            symbol (str): Symbol to check. Defaults to ''.
        """
        positions = await Positions().positions_get(symbol=symbol)
        loosing = [trade for trade in positions if trade.profit <= 0]
        return len(loosing) >= self.loss_limit

    async def check_open_positions(self, *, symbol: str = '') -> bool:
        """Check if the number of open positions is greater than or equal the loss limit.

        Args:
            symbol (str): Symbol to check. Defaults to ''.
        """
        positions = await Positions().positions_get(symbol=symbol)
        return len(positions) >= self.open_limit

    async def check_risk_level(self) -> bool:
        """Check the risk level."""
        await self.account.refresh()
        risk_level = (1 - (self.account.margin_free / self.account.equity)) * 100
        return risk_level >= self.risk_level



from functools import cache
import enum
from logging import getLogger

from .config import Config
from .meta_trader import MetaTrader

logger = getLogger(__name__)


class Base:
    """A base class for all data structure classes in the aiomql package. This class provides a set of common methods
     and attributes for handling data.
    """
    mt5: MetaTrader
    config: Config

    def __init__(self, **kwargs):
        """
        Initialize a new instance of the Base class
        Args:
            **kwargs: Set instance attributes with keyword arguments. Only if they are annotated on the class body.
        """
        self.config = Config()
        self.mt5 = MetaTrader()
        self.exclude = {'mt5', "config", 'exclude', 'include', 'annotations', 'class_vars', 'dict'}
        self.include = set()
        self.set_attributes(**kwargs)

    def __repr__(self):
        kv = [(k, v) for k, v in self.__dict__.items() if not k.startswith('_') and
              (type(v) in (int, float, str) or isinstance(v, enum.Enum))]
        args = (', '.join('%s=%s' % (i, j) for i, j in kv[:3]))
        args = args if len(kv) <= 3 else args + ' ... ' + ', '.join('%s=%s' % (i, j) for i, j in kv[-1:])
        return '%(class)s(%(args)s)' % {'class': self.__class__.__name__, 'args': args}

    def set_attributes(self, **kwargs):
        """Set keyword arguments as object attributes
        
        Keyword Args:
            **kwargs: Object attributes and values as keyword arguments

        Raises:
            AttributeError: When assigning an attribute that does not belong to the class or any parent class

        Notes:
            Only sets attributes that have been annotated on the class body.
        """
        for i, j in kwargs.items():
            try:
                setattr(self, i, self.annotations[i](j))
            except KeyError:
                logger.warning(f"Attribute {i} does not belong to class {self.__class__.__name__}")
                continue

            except ValueError:
                logger.warning(f'Cannot covert object of type {type(j)} to type {self.annotations[i]}')
                continue

            except Exception as exe:
                logger.warning(f'Did not set attribute {i} on class {self.__class__.__name__} due to {exe}')
                continue

    @property
    @cache
    def annotations(self) -> dict:
        """Class annotations from all ancestor classes and the current class.

        Returns:
            dict: A dictionary of class annotations
        """
        annots = {}
        for base in self.__class__.__mro__[-3::-1]:
            annots |= getattr(base, '__annotations__', {})
        return annots

    def get_dict(self, exclude: set = None, include: set = None) -> dict:
        """Returns class attributes as a dict, with the ability to filter

        Keyword Args:
            exclude: A set of attributes to be excluded
            include: Specific attributes to be returned

        Returns:
             dict: A dictionary of specified class attributes

        Notes:
            You can only set either of include or exclude. If you set both, include will take precedence
        """
        exclude, include = exclude or set(), include or set()
        filter_ = include or set(self.dict.keys()).difference(exclude)
        return {key: value for key, value in self.dict.items() if key in filter_}

    @property
    @cache
    def class_vars(self):
        """Annotated class attributes

        Returns:
            dict: A dictionary of available class attributes in all ancestor classes and the current class.
        """
        clss = self.__class__.__mro__[-3::-1]
        cls_dict = {}
        for cls in clss:
            cls_dict |= cls.__dict__
        return {key: value for key, value in cls_dict.items() if key in self.annotations}

    @property
    def dict(self) -> dict:
        """All instance and class attributes as a dictionary, except those excluded in the Meta class.

        Returns:
            dict: A dictionary of instance and class attributes
        """
        try:
            _filter = self.exclude.difference(self.include)
            return {key: value for key, value in (self.class_vars | self.__dict__).items() if
                    key not in _filter}
        except Exception as err:
            logger.warning(err)



import asyncio
from logging import getLogger

from .core.models import TradeRequest, OrderSendResult, OrderCheckResult, TradeOrder
from .core.constants import TradeAction, OrderTime, OrderFilling
from .core.exceptions import OrderError
logger = getLogger(__name__)


class Order(TradeRequest):
    """Trade order related functions and properties. Subclass of TradeRequest."""

    def __init__(self, **kwargs):
        """Initialize the order object with keyword arguments, symbol must be provided.
        Provide default values for action, type_time and type_filling if not provided.

        Args:
            **kwargs: Keyword arguments must match the attributes of TradeRequest as well as the attributes of
             Order class as specified in the annotations in the class definition.

        Default Values:
            action (TradeAction.DEAL): Trade action
            type_time (OrderTime.DAY): Order time
            type_filling (OrderFilling.FOK): Order filling
        """
        if 'symbol' in kwargs:
            kwargs['symbol'] = str(kwargs['symbol'])
        self.action = kwargs.pop('action', TradeAction.DEAL)
        self.type_time = kwargs.pop('type_time', OrderTime.DAY)
        self.type_filling = kwargs.pop('type_filling', OrderFilling.FOK)
        super().__init__(**kwargs)

    async def orders_total(self):
        """Get the number of active orders.

        Returns:
            (int): total number of active orders
        """
        return await self.mt5.orders_total()

    async def get_order(self, *, ticket: int, retries: int = 3) -> TradeOrder | None:
        """
        Get the order by ticket number.
        Args:
            ticket (int): Order ticket number
            retries (int): Number of retries
        Returns:
        """
        if retries < 1:
            return None
        orders = await self.mt5.orders_get(ticket=ticket)
        if orders and (order := orders[0]).ticket == ticket:
            return TradeOrder(**order._asdict())
        if self.mt5.error.is_connection_error():
            await asyncio.sleep(retries)
            return await self.get_order(ticket=ticket, retries=retries-1)
        return None

    async def get_orders(self, *, ticket: int = 0, symbol: str = '', group: str = '', retries=3)\
            -> tuple[TradeOrder, ...]:
        """Get the list of active orders for the current symbol.
        Keyword Args:
            ticket (int): Order ticket number
            symbol (str): Symbol name
            group (str): Group name
        Returns:
            tuple[TradeOrder]: A Tuple of active trade orders as TradeOrder objects
        """
        if retries < 1:
            return tuple()
        symbol = getattr(self, 'symbol', symbol)
        orders = await self.mt5.orders_get(symbol=symbol, ticket=ticket, group=group)
        if orders is not None:
            orders = (TradeOrder(**order._asdict()) for order in orders)
            return tuple(orders)
        if self.mt5.error.is_connection_error():
            await asyncio.sleep(retries)
            return await self.get_orders(ticket=ticket, symbol=symbol, group=group, retries=retries-1)
        return tuple()

    async def check(self, **kwargs) -> OrderCheckResult:
        """Check funds sufficiency for performing a required trading operation and the possibility of executing it.

        Returns:
            OrderCheckResult: An OrderCheckResult object

        Raises:
            OrderError: If not successful
        """
        req = self.dict | kwargs
        res = await self.mt5.order_check(req)
        if res is None:
            raise OrderError(f'Failed to check order due to {self.mt5.error.description}')
        return OrderCheckResult(**res._asdict())

    async def send(self) -> OrderSendResult:
        """Send a request to perform a trading operation from the terminal to the trade server.

        Returns:
             OrderSendResult: An OrderSendResult object

        Raises:
            OrderError: If not successful
        """
        res = await self.mt5.order_send(self.dict)
        if res is None:
            raise OrderError(f'Failed to send order {self.symbol} due to {self.mt5.error.description}')
        res = OrderSendResult(**res._asdict())
        try:
            profit = await self.calc_profit()
            loss = await self.calc_profit(tp=self.sl)
            res.loss = loss
            res.profit = profit
        except Exception as exe:
            logger.error(f'Failed to calculate profit and loss for this order due to {exe}')
        return res

    async def calc_margin(self) -> float:
        """Return the required margin in the account currency to perform a specified trading operation.

        Returns:
            float: Returns float value if successful

        Raises:
            OrderError: If not successful
        """
        res = await self.mt5.order_calc_margin(self.type, self.symbol, self.volume, self.price)
        if res is None:
            raise OrderError(f'Failed to calculate margin for {self.symbol} due to {self.mt5.error.description}')
        return res

    async def calc_profit(self, **kwargs) -> float:
        """Return profit in the account currency for a specified trading operation.

        Returns:
            float: Returns float value if successful

        Raises:
            OrderError: If not successful
        """
        args = self.get_dict(include={'tp', 'price', 'symbol', 'volume', 'type'})
        args |= kwargs
        res = await self.mt5.order_calc_profit(args['type'], args['symbol'], args['volume'], args['price'], args['tp'])
        if res is None:
            raise OrderError(f'Failed to calculate profit for {self.symbol} due to {self.mt5.error.description}')
        return res



"""Module for working with price ticks."""

from typing import TypeVar, Iterable

from pandas import DataFrame, Series
import pandas_ta as ta
import mplfinance as mplt
import pandas as pd

from .core.constants import TickFlag

Self = TypeVar('Self', bound='Ticks')


class Tick:
    """Price Tick of a Financial Instrument.

    Attributes:
        time (int): Time of the last prices update for the symbol
        bid (float): Current Bid price
        ask (float): Current Ask price
        last (float): Price of the last deal (Last)
        volume (float): Volume for the current Last price
        time_msc (int): Time of the last prices update for the symbol in milliseconds
        flags (TickFlag): Tick flags
        volume_real (float): Volume for the current Last price
        Index (int): Custom attribute representing the position of the tick in a sequence.
    """
    time: float
    bid: float
    ask: float
    last: float
    volume: float
    time_msc: float
    flags: float
    volume_real: float
    Index: int

    def __init__(self, **kwargs):
        """Initialize the Tick class. Set attributes from keyword arguments. bid, ask, last, time and volume must be
        present"""
        if not all(key in kwargs for key in ['bid', 'ask', 'last', 'volume', 'time']):
            raise ValueError("bid, ask, last and volume, time must be present in the keyword arguments")
        self.Index = kwargs.pop('Index', 0)
        self.set_attributes(**kwargs)

    def __repr__(self):
        return ("%(class)s(Index=%(Index)s, time=%(time)s, bid=%(bid)s, ask=%(ask)s, last=%(last)s, volume=%(volume)s)"
                % {"class": self.__class__.__name__, "time": self.time, "bid": self.bid,
                   "ask": self.ask, "last": self.last, "volume": self.volume, 'Index': self.Index})

    def dict(self, exclude: set = None, include: set = None) -> dict:
        """
        Returns a dictionary of the instance attributes.

        Args:
            exclude: A set of attributes to exclude from the dictionary. Defaults to None.
            include: A set of attributes to include in the dictionary. Defaults to None.

        Returns: dict
        """
        exclude = exclude or set()
        include = include or set()
        keys = include or set(self.__dict__.keys()).difference(exclude)
        return {k: v for k, v in self.__dict__.items() if k in keys}

    def set_attributes(self, **kwargs):
        """Set attributes from keyword arguments"""
        for key, value in kwargs.items():
            setattr(self, key, value)


_Ticks = TypeVar('_Ticks', bound='Ticks')


class Ticks:
    """Container class for price ticks. Arrange in chronological order. Supports iteration, slicing and assignment"""
    time: Series
    bid: Series
    ask: Series
    last: Series
    volume: Series
    time_msc: Series
    flags: Series
    volume_real: Series
    Index: Series

    def __init__(self, *, data: DataFrame | Iterable, flip=False):
        """Initialize the Ticks class. Creates a DataFrame of price ticks from the data argument.

        Args:
            data (DataFrame | Iterable): Dataframe of price ticks or any iterable object that can be converted to a
                pandas DataFrame
            flip (bool): If flip is True reverse data chronological order.
        """
        if isinstance(data, DataFrame):
            data = data
        elif isinstance(data, type(self)):
            data = DataFrame(data.data)
        elif isinstance(data, Iterable):
            data = DataFrame(data)
        else:
            raise ValueError(f'Cannot create DataFrame from object of {type(data)}')
        self._data = data.iloc[::-1] if flip else data

    def __repr__(self):
        return self._data.__repr__()

    def __len__(self):
        return self._data.shape[0]

    def __contains__(self, item: Tick) -> bool:
        return item.time_msc == self[item.Index].time_msc

    def __getattr__(self, item):
        if item in list(self._data.columns.values):
            return self._data[item]
        raise AttributeError(f'Attribute {item} not defined on class {self.__class__.__name__}')

    def __getitem__(self, index) -> Tick | Self:
        if isinstance(index, slice):
            cls = self.__class__
            data = self._data.iloc[index]
            data.reset_index(drop=True, inplace=True)
            return cls(data=data)

        if isinstance(index, str):
            return self._data[index]

        item = self._data.iloc[index]
        return Tick(**item, Index=index)

    def __setitem__(self, index, value: Series):
        if isinstance(value, Series):
            self._data[index] = value
            return
        raise TypeError(f'Expected Series got {type(value)}')

    def __iter__(self):
        return (Tick(**row._asdict()) for row in self._data.itertuples())

    @property
    def ta(self):
        """Access to the pandas_ta library for performing technical analysis on the underlying data attribute.

        Returns:
            pandas_ta: The pandas_ta library
        """
        return self._data.ta

    @property
    def ta_lib(self):
        """Access to the ta library for performing technical analysis. Not dependent on the underlying data attribute.

        Returns:
            ta: The ta library
            """
        return ta

    @property
    def data(self) -> DataFrame:
        """DataFrame of price ticks arranged in chronological order."""
        return self._data

    def rename(self, inplace=True, **kwargs) -> _Ticks | None:
        """Rename columns of the candle class.

        Keyword Args:
            inplace (bool): Rename the columns inplace or return a new instance of the class with the renamed columns
            **kwargs: The new names of the columns

        Returns:
            Ticks: A new instance of the class with the renamed columns if inplace is False.
            None: If inplace is True
        """
        res = self._data.rename(columns=kwargs, inplace=inplace)
        return res if inplace else self.__class__(data=res)

    def make_addplot(self, *, count: int = 50, columns: list = None, **kwargs) -> dict:
        """
        Make subplots for adding to the main plot

        Args:
            count (int): The numbers of candles to make the addplot for. Defaults to 50.
            columns (list[str]): The columns to make the plot from. Defaults to None.
            **kwargs: Valid arguments for the mplfinance make_addplot function
        """
        columns = columns or []
        data = self._data[-count:]
        data.index = pd.to_datetime(data['time'], unit='s')
        return mplt.make_addplot(data[columns], **kwargs)

    def visualize(self, *, count: int = 50, type='candle', savefig: str | dict = None, addplot: dict = None,
                  style: str = 'charles', ylabel: str = 'Price', title: str = 'Chart', **kwargs):
        """Visualize the candles using the mplfinance library.
        Args:
            count (int): The number of candles to visualize, counting from behind, i.e the most recent candles.
             Defaults to 50.
            type: Type of chart, defaults to candle
            savefig (str|dict): The path to save the figure or a dictionary of parameters to pass to the savefig method.
            addplot: Additional plots to add to the chart. Defaults to None. They should match the dimension of the
              original data which is specified via the count parameter.
            style (str): The style of the chart. Defaults to 'charles'.
            ylabel (str): The label of the y-axis. Defaults to 'Price'.
            title (str): The title of the chart. Defaults to 'Chart'.
            kwargs: valid kwargs for the plot function.
        """
        kwargs |= {key: arg for key, arg in (('savefig', savefig), ('addplot', addplot), ('style', style),
                                             ('ylabel', ylabel), ('title', title), ('type', type)) if arg}
        data = self._data[-count:]
        data.index = pd.to_datetime(data['time'], unit='s')
        mplt.plot(data, **kwargs)



import csv
import json
from logging import getLogger
from typing import Iterable, Literal

from .core import Config
from .core.models import OrderSendResult

logger = getLogger(__name__)


class Result:
    """A base class for handling trade results and strategy parameters for record keeping and reference purpose.
    The data property must be implemented in the subclass

    Attributes:
        config (Config): The configuration object
        name: Any desired name for the result file object
    """
    config: Config

    def __init__(self, result: OrderSendResult, parameters: dict = None, name: str = ''):
        """
        Prepare result data
        Args:
            result:
            parameters:
            name:
        """
        self.config = Config()
        self.parameters = parameters or {}
        self.result = result
        self.name = name or parameters.get('name', 'Trades')

    def get_data(self) -> dict:
        res = self.result.get_dict(exclude={'retcode', 'comment', 'retcode_external', 'request_id', 'request'})
        return self.parameters | res | {'actual_profit': 0, 'closed': False, 'win': False}

    async def save(self, *, trade_record_mode: Literal['csv', 'json'] = None):
        """Record trade results as a csv or json file
        Args:
            trade_record_mode (Literal['csv'|'json']): Mode of saving trade records
        """
        trade_record_mode = trade_record_mode or self.config.trade_record_mode
        if trade_record_mode == 'csv':
            await self.to_csv()
        else:
            await self.to_json()

    async def to_csv(self):
        """Record trade results and associated parameters as a csv file
        """
        try:
            data = self.get_data()
            file = self.config.records_dir / f"{self.name}.csv"
            file.touch(exist_ok=True) if not file.exists() else ...
            reader: Iterable[dict] = csv.DictReader(file.open('r', newline=''))
            rows: list[dict] = []
            headers = set()
            [(rows.append(row), headers.update(row.keys())) for row in reader]
            rows.append(data)
            headers.update(data.keys())
            writer = csv.DictWriter(file.open('w', newline=''), fieldnames=headers, restval=None,
                                    extrasaction='ignore')
            writer.writeheader()
            writer.writerows(rows)
        except Exception as err:
            logger.error(f'Unable to save to csv: {err}')

    @staticmethod
    def serialize(value) -> str:
        """Serialize the trade records and strategy parameters
        """
        try:
            return str(value)
        except (ValueError, TypeError) as _:
            return ""

    async def to_json(self):
        """Save trades and strategy parameters in a json file
        """
        try:
            file = self.config.records_dir / f"{self.name}.json"
            data = self.get_data()
            exists = file.touch(exist_ok=True) if not file.exists() else True
            if not exists:
                json.dump([], file.open('w'))
            with file.open('r') as fh:
                rows = json.load(fh)
            rows.append(data)
            with file.open('w') as fh:
                json.dump(rows, fh, indent=2, skipkeys=True, default=self.serialize)
        except Exception as err:
            logger.error(f"Unable to save as json file: {err}")
